 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

 /*
  * $Log$
  */

/*
 * Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Main program for kmsquish
   >>>> 
   >>>>  Private: 
   >>>> 	main
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "kmsquish.h"

clui_info_struct *clui_info = NULL;

/*-----------------------------------------------------------
|
|  Routine Name: main() - Compress Map to One Column by Means of Average, RMS, or MAX
|
|       Purpose: main program for kmsquish
|
|         Input:
|		char *clui_info->i_file; {First Input data object}
|		int   clui_info->i_flag; {TRUE if -i specified}
|
|		char *clui_info->o_file; {Resulting output data object}
|		int   clui_info->o_flag; {TRUE if -o specified}
|
|		int clui_info->map_logic; {map data after squishing map (1) or leave map with only one column (0)}
|		int clui_info->map_flag; {TRUE if -map specified}
|
|		int clui_info->t_toggle; {squishing scheme (1)average, (2)RMS, (3)Maximum}
|		int clui_info->t_flag; {TRUE if -t specified}
|
|        Output:
|       Returns:
|
|    Written By: Scott Wilson (stolen from K1.5 vmsquish, which was written by  Tom Sauer)
|          Date: Apr 08, 1995
| Modifications:
|
------------------------------------------------------------*/

void main(
   int  argc,
   char **argv,
   char **envp)
{

	kform  *pane;         /* form tree representing *.pane file */
/* -main_variable_list */
        kobject in_obj = NULL;
        kobject out_obj = NULL;
/* -main_variable_list_end */

	khoros_initialize(argc, argv, envp, "DATAMANIP");
	kexit_handler(kmsquish_free_args, NULL);

/* -main_get_args_call */
	pane = kgen_initialize(PANEPATH, KGEN_KROUTINE, "DATAMANIP", "kmsquish",
		kmsquish_usage_additions);

	if (!(kclui_check_args()))
	    kexit(KEXIT_FAILURE);
	kmsquish_get_args(pane);
	kvf_destroy_form(pane);
/* -main_get_args_call_end */

/* -main_before_lib_call */
        if ((in_obj = kpds_open_input_object(clui_info->i_file))
                == KOBJECT_INVALID)
        {
           kerror("kmsquish","main", "Cannot open input object.");
           kexit(KEXIT_FAILURE);
        }
 
        if ((out_obj = kpds_open_output_object(clui_info->o_file))
                == KOBJECT_INVALID)
        {
           kerror("kmsquish","main", "Cannot open output object.");
           kexit(KEXIT_FAILURE);
        }

        if (!kpds_copy_object_attr(in_obj, out_obj))
        {
           kerror("kmsquish","main",
                  "Unable to copy attributes from input to output object.");
           kexit(KEXIT_FAILURE);
        }
/* -main_before_lib_call_end */

/* -main_library_call */
        if (!lkmsquish(in_obj,
                      clui_info->map_logic,
                      clui_info->t_toggle,
                      out_obj))
                 kexit(KEXIT_FAILURE);
/* -main_library_call_end */

/* -main_after_lib_call */
        /* add history to the output object */
        if (!kpds_set_attribute(out_obj, KPDS_HISTORY, kpds_history_string()))
        {
           kerror("kmsquish", "main",
                  "Unable to set history on the destination object");
            kexit(KEXIT_FAILURE);
        }
 
        kpds_close_object(in_obj);
        kpds_close_object(out_obj);
/* -main_after_lib_call_end */


	kexit(KEXIT_SUCCESS);
}


/*-----------------------------------------------------------
| 
|  Routine Name: kmsquish_usage_additions
| 
|       Purpose: Prints usage additions in kmsquish_usage routine
| 
|         Input: None
| 
|        Output: None
|    Written By: ghostwriter -oname kmsquish
|          Date: Apr 08, 1995
| Modifications: 
| 
------------------------------------------------------------*/
void kmsquish_usage_additions(void)
{
	kfprintf(kstderr, "\tCompress Map to One Column by Means of Average, RMS, or MAX\n");

/* -usage_additions */
/* -usage_additions_end */

}
/*-----------------------------------------------------------
| 
|  Routine Name: kmsquish_free_args
| 
|       Purpose: Frees CLUI struct allocated in kmsquish_get_args()
| 
|         Input: None
| 
|        Output: None
|    Written By: ghostwriter -oname kmsquish
|          Date: Apr 08, 1995
| Modifications: 
| 
------------------------------------------------------------*/
/* ARGSUSED */
void
kmsquish_free_args(
    int   status,
    kaddr client_data)
{

	/* do the wild and free thing */
	if (clui_info != NULL)
		{
	kfree(clui_info->i_file);
	kfree(clui_info->o_file);
		kfree(clui_info);
		}

/* -free_handler_additions */
/* -free_handler_additions_end */
}
